/*
 *  PEPTO.C
 *
 *  Fast, soothing relief for C source files suffering
 *  from acid indigestion and that bloated feeling.
 *
 *  Compile with Turbo C 2.0, small memory model.
 *
 *  David Stafford (CIS 72411,2670)
 *
 *  4/8/89    1.0
 *
 *  4/28/89   1.1 fixed bug where a space is needed as a
 *            separator, between a forward slash and asterisk
 *            as in  temp = 99 / *ptr;
 *            Added wildcard file name support.
 *
 *  5/1/89    1.2 fixed minor bug introduced in 1.1 where
 *            an extra CR/LF was generated unnecessarily.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dir.h>

#define MAXLEN          1200              /* maximum length for an identifier */
#define NUM_CHARS       256

#define IDENTIFIER      1                 /* some token types */
#define CR              2
#define FORWARD_SLASH   3

#define FALSE           0
#define TRUE            !FALSE

/* Pepto is lexically driven.  Each ASCII character has an associated        */
/* function which interprets that character.  For example, the forward slash */
/* '/' is associated with process_forward_slash() which looks to see if the  */
/* next character is an asterisk.  To manage this we keep an array of        */
/* pointers to functions which is indexed by each character.                 */
void (*character_table[ NUM_CHARS ])( int );

FILE *fpin, *fpout;                       /* input and output files */
int in_column, last_in_column;
int out_column;                           /* current output column */
char ident[ MAXLEN + 1 ];                 /* identifier buffer */
int last_token;                           /* last token type read */


/*-------------------- output primitive -----------------------*/
void put( int c )
  {
  last_token = NULL;

  if( c == '\n' )  out_column = 0;
  else             out_column = 1;

  putc( c, fpout );
  }


/*-------------------- input primitives -----------------------*/
int get( void )
  {
  int c;

  last_in_column = in_column;

  c = getc( fpin );

  if( c == '\n' )  in_column = 0;
  else             in_column = 1;

  return( c );
  }


int peek( void )
  {
  int c;

  c = getc( fpin );
  ungetc( c, fpin );

  return( c );
  }


/*-------------------- backup primitive -----------------------*/
void unget( int c )
  {
  ungetc( c, fpin );
  in_column = last_in_column;
  }


/*--------------------- state actions --------------------------*/
void whitespace( int c )
  {
  if( last_in_column == 0 )  in_column = 0;
  }


void copy( int c )
  {
  put( c );
  }


void process_cr( int c )
  {
  if( last_token != CR )
    {
    put( '\n' );
    last_token = CR;
    }
  }


void process_asterisk( int c )
  {
  if( last_token == FORWARD_SLASH )     /* don't merge them, it's a comment! */
    {
    put( ' ' );
    last_token = NULL;
    }

  put( '*' );
  }


void process_digit( int c )
  {
  if( last_token == IDENTIFIER )  put( ' ' );
  put( c );
  }


void process_escape_character( int c )
  {
  put( '\\' );
  put( get() );
  }


void process_character_constant( int c )
  {
  put( '\'' );

  while( (c = get()) != EOF )
    {
    if( c == '\\' )  process_escape_character( 0 );
    else             put( c );

    if( c == '\'' )  return;
    }
  }


void process_string_constant( int c )
  {
  put( '"' );

  while( (c = get()) != EOF )
    {
    if( c == '\\' )  process_escape_character( 0 );
    else             put( c );

    if( c == '"' )  return;
    }
  }


void process_comment( int c )
  {
  while( (c = get()) != EOF )
    {
    if( c == '*' && peek() == '/' )
      {
      get();
      return;
      }
    }
  }


void process_forward_slash( int c )
  {
  if( peek() == '*' )
    {
    get();
    process_comment( 0 );
    }
  else
    {
    put( '/' );
    last_token = FORWARD_SLASH;
    }
  }


void process_backward_slash( int c )
  {
  if( peek() == '\n' )
    {
    get();
    last_in_column = in_column = 1;
    }
  else
    {
    put( '\\' );
    }
  }


void strip_spaces( void )
  {
  int c;

  while( c = peek(), c == ' ' || c == '\t' )
    {
    get();
    }
  }


void process_identifier( int c )
  {
  int i, len = 1, col = last_in_column;

  if( last_token == IDENTIFIER )  put( ' ' );

  ident[ 0 ] = c;

  while( c = peek(), c != EOF && (c == '_' || isalnum( c )) )
    {
    get();
    if( len < MAXLEN )  ident[ len++ ] = c;
    }

  ident[ len ] = 0;

  /* make sure 'asm' keyword appears on its own line if necessary */
  if( !stricmp( ident, "asm" ) && out_column != 0 && col == 0 )
    {
    put( '\n' );
    }

  for( i = 0; i < len; ++i )
    {
    put( ident[ i ] );
    }

  last_token = IDENTIFIER;
  }


void process_define( void )
  {
  int c, paren = 1;

  strip_spaces();
  put( ' ' );                   /* put a space after "#define" */

  while( c = get(), c != EOF && c != ' ' && c != '\t' && c != '\n' && c != '(' )
    {
    put( c );
    }

  if( c == EOF )  return;

  if( c == '\n' )
    {
    put( '\n' );
    last_token = CR;
    return;
    }

  if( c == ' ' || c == '\t' )  put( ' ' );

  if( c == '(' )
    {
    put( '(' );

    while( paren && (c = get()) != EOF )
      {
      (*character_table[ c ])( c );
      if( c == '(' )  ++paren;
      if( c == ')' )  --paren;
      }

    put( ' ' );
    }

  while( (c = get()) != EOF )
    {
    if( c == '\n' )
      {
      put( '\n' );
      last_token = CR;
      return;
      }

    (*character_table[ c ])( c );
    }
  }


void process_preprocessor( int c )
  {
  if( out_column != 0 )  put( '\n' );

  put( '#' );
  strip_spaces();

  c = get();

  if( c == 'd' || c == 'D' )     /* looking for #define */
    {
    process_identifier( c );

    if( !stricmp( ident, "define" ) )
      {
      process_define();
      return;
      }
    }

  while( c != EOF )
    {
    if( c == '\n' )
      {
      put( '\n' );
      last_token = CR;
      return;
      }

    (*character_table[ c ])( c );

    c = get();
    }
  }


void process_pound( int c )
  {
  if( last_in_column == 0 )
    {
    process_preprocessor( 0 );
    }
  else
    {
    put( '#' );
    }
  }


/*---------------------------------------------*/
void init_character_table( int cat )
  {
  int i;

  for( i = 0; i < NUM_CHARS; ++i )
    {
    character_table[ i ] = copy;
    }

  if( cat )
    {
    character_table[ '\\' ] = process_backward_slash;
    character_table[ '\n' ] = whitespace;
    }
  else
    {
    character_table[ '\n' ] = process_cr;
    }

  character_table[ '"'  ] = process_string_constant;
  character_table[ '\'' ] = process_character_constant;
  character_table[ '/'  ] = process_forward_slash;
  character_table[ '#'  ] = process_pound;
  character_table[ ' '  ] = whitespace;
  character_table[ '\t' ] = whitespace;
  character_table[ '\r' ] = whitespace;
  character_table[ '_'  ] = process_identifier;
  character_table[ '*'  ] = process_asterisk;

  for( i = 0; i < 26; ++i )
    {
    character_table[ 'a' + i ] = process_identifier;
    character_table[ 'A' + i ] = process_identifier;
    }

  for( i = '0'; i <= '9'; ++i )
    {
    character_table[ i ] = process_digit;
    }
  }


/*------------ steps though the source file ---------------*/
void bismol( void )
  {
  int c;

  while( (c = get()) != EOF )
    {
    (*character_table[ c ])( c );
    }
  }


/*------------------- prepares and cleans up files ----------------------*/
void pepto( char *fin )
  {
  char fout[ MAXPATH ], *dot;

  /* change filename suffix to .PEP for output file */
  strcpy( fout, fin );
  if( (dot = strchr( fout, '.' )) == NULL )  dot = fout + strlen( fout );
  strcpy( dot, ".PEP" );

  if( !stricmp( fin, fout ) )  return;     /* do not PEPTO .PEP files! */

  if( (fpin = fopen( fin, "r" )) == NULL )
    {
    fputs( "Error: Can't open ", stdout );
    puts( fin );
    return;
    }

  puts( fin );

  /* allocate a 16k input buffer */
  if( setvbuf( fpin, NULL, _IOFBF, 16 * 1024 ) )
    {
    puts( "Error: Not enough RAM." );
    }

  if( (fpout = fopen( fout, "w" )) == NULL )
    {
    fputs( "Error: Can't open ", stdout );
    puts( fout );
    return;
    }

  /* allocate a 16k output buffer */
  if( setvbuf( fpout, NULL, _IOFBF, 16 * 1024 ) )
    {
    puts( "Error: Not enough RAM." );
    }

  /* init globals */
  in_column = out_column = 0;
  ident[ 0 ] = '\0';
  last_token = CR;

  bismol();

  fclose( fpin );
  fclose( fpout );
  }


/*-------------------------------------------------------------------*/
void main( int argc, char *argv[] )
  {
  char drive[ MAXDRIVE ], dir[ MAXDIR ], file[ MAXFILE ], ext[ MAXEXT ];
  char path[ MAXPATH ];
  struct ffblk fb;
  int i, cat = TRUE;

  puts( "Pepto 1.2 - David Stafford (CIS 72411,2670)\n" );

  if( argc == 1 )
    {
    puts( "Provides fast, soothing relief for C source files" );
    puts( "suffering from acid indigestion and that bloated feeling.\n" );
    puts( "Usage: PEPTO </C> filename <filename>" );
    puts( "       /C prevents concatenation of source lines" );
    puts( "       Output file suffix is .PEP" );
    puts( "       Wildcards OK" );
    }
  else
    {
    while( --argc )
      {
      strupr( *++argv );

      if( !strcmp( *argv, "/C" ) || !strcmp( *argv, "-C" ) )
        {
        cat = FALSE;
        }
      else
        {
        init_character_table( cat );

        fnsplit( *argv, drive, dir, file, ext );

        for( i = findfirst( *argv, &fb, 0 ); !i; i = findnext( &fb ) )
          {
          fnmerge( path, drive, dir, NULL, NULL );
          strcat( path, fb.ff_name );
          pepto( path );
          }
        }
      }
    }
  }

